#!/usr/bin/env python3

# tower.py - Tower of Mystery text adventure game

# The original BASIC version of this game and backround information can
# be found at https://www.jasonawells.org/tower. This port to Python 3
# provides commentary on how the source works. Many of the original BASIC
# algorithms are not appropriate in Python 3, so there is some attempt to
# update the code to an object-oriented model.

# Like many text adventure games, the main loop of the game describes
# the player's location, asks for a command in the form "VERB NOUN" eg
# "GO NORTH" or "TAKE COIN," parses the command into its constituent words
# and displays the results of the command.

# At the bottom of the file is the main game loop which gives the overview
# of the game logic. Search for MAIN GAME LOOP to see those steps and to
# follow the logic.

# Directions are a special class of verb. The full table of verbs ("Verbs") is below.

Directions = {
	"N": [0, "NORTH"],
	"S": [1, "SOUTH"], 
	"E": [2, "EAST"],
	"W": [3, "WEST"],
	"U": [4, "UP"],
	"D": [5, "DOWN"]
}

# The game has three data types in the classes:
# 1. Player
# 2. Thing
# 3. Room

# 1. The Player has an inventory list, a location on the map (room) and
# flags for game state. The Player also has a "counter" used to time
# how many turns the player has to use the in-game computer object. The
# "wantDescription" flag determines whether or not so show the current
# room description.

class Player:
	def __init__(self):
		self.inventory = []
		self.room = 1
		self.counter = 0	
		self.wantDescription = True
		self.coinFound = False
		self.ratsAsleep = False
		self.computerActive = False
		self.manualFound = False
		self.computerDead = False
		self.loggedIn = False
		self.bats = True
		self.mountedTape = False
		self.copyRunning = False

# 2. Every Thing in the game has a text description and a flag if it can
# be picked up.  E.g., a coin can be taken but the player cannot pick up
# the hungry rats. The location of a Thing is not stored. Things will be
# added to the contents list of Rooms (see below). If a Thing is takable
# then it can be added to the inventory list of the Player (see above).

class Thing:
	def __init__(self, description, takable):
		self.description = description
		self.takable = takable

# 3. A Room has a description string, a list of contents and a list of exits
# in various directions. The exits list is in the order of the Directions list
# above (N, S, E, W, U, D). Each Room has a method to describe itself.

class Room:
	def __init__(self, description, contents, exits):
		self.description = description
		self.contents = contents
		self.exits = exits

	def describe(self):
		print(self.description)

		if len(self.contents) > 0:
			print("You see:")
			for thing in self.contents:
				print(thing.description)

		print("Exits are:")
		for direction in Directions:
			if self.exits[Directions[direction][0]] != 0:
				print(direction, end=" ")

		print()

# With the basic classes above, we can create the actual game objects.
# If you leave most things above as they are, you can code a new game
# from the code below.

# The Things dictionary creates the game objects with a description and
# a takable flag. The dictionary entries are 5 letter, all-caps because
# of the original BASIC source code. This Python code will convert entered
# words into this format for lookup.

Things = {
	"RATS":		Thing("Hungry rats", False),
	"TAPE":		Thing("Computer tape", True), 
	"MACHI":	Thing("Vending machine", False), 
	"TERMI":	Thing("Broken-down terminal", True), 
	"COIN":		Thing("Coin", True), 
	"CANDY":	Thing("Candy Bar", True), 
	"COMPU":	Thing("Computer", False), 
	"BATS":		Thing("Bats", False), 
	"DESK":		Thing("Desk", False), 
	"MANUA":	Thing("Computer Manual", True), 
	"CLOCK":	Thing("Elaborate Clockwork", False), 

	# pseudo-objects - these are not Things like the above objects
	# that can be manipulated in the game. The original BASIC source
	# treated these tokens as pseudo-objects for the commands TYPE and GO
	# (e.g., "TYPE ROAD" or "GO NORTH") where these tokens are treated
	# as direct objects.

	"ROAD": 1,
	"DIR": 2,
	"ADVEN": 3,
	"COPY": 4,
	"LOGOU": 5,
	"NORTH": 6,
	"SOUTH": 7,
	"EAST": 8,
	"WEST": 9,
	"UP": 10,
	"DOWN": 11,
}

# The Rooms list is a full map of the game. It is a list of rooms with a description,
# a list of Things in each room and a list of exits to other rooms. The room numbers
# in the original BASIC game were 1-based and not 0-based, so this port
# adds a "You are nowhere" room as room 0. It should not be possible to
# reach this room according to game rules.

Rooms = []
Rooms.append(Room("You are nowhere.", [], [0, 0, 0, 0, 0, 0]))
Rooms.append(Room("You are in front of an old factory with a clock tower.", [], [2, 0, 0, 0, 0, 0]))
Rooms.append(Room("You are at the bottom of the stairwell.", [], [3, 1, 0, 0, 7, 0]))
Rooms.append(Room("You are at the top of some basement steps.", [], [0, 2, 0, 0, 0, 4]))
Rooms.append(Room("You are in a damp cellar.", [Things["RATS"]], [0, 0, 5, 0, 3, 0]))
Rooms.append(Room("You are in a storeroom.", [Things["TAPE"]], [0, 0, 0, 4, 0, 0]))
Rooms.append(Room("You are in the cafeteria.", [Things["MACHI"]], [0, 0, 7, 0, 0, 0]))
Rooms.append(Room("You're at a landing on the stairs.", [], [0, 0, 8, 6, 9, 2]))
Rooms.append(Room("Around you is a manufacturing area.", [Things["TERMI"]], [0, 0, 0, 7, 0, 0]))
Rooms.append(Room("You're at a landing on the third floor.", [], [0, 0, 10, 0, 12, 7]))
Rooms.append(Room("You are in the computer room.", [Things["COMPU"]], [0, 0, 0, 9, 0, 0]))
Rooms.append(Room("You are inside the clock tower.", [Things["CLOCK"]], [0, 12, 0, 0, 0, 0]))
Rooms.append(Room("You're at the top of the stairs.", [], [11, 0, 13, 0, 0, 9]))
Rooms.append(Room("You are in a long corridor going east.", [Things["BATS"]], [0, 0, 14, 12, 0, 0]))
Rooms.append(Room("You're at the east end of the corridor.", [Things["DESK"]], [0, 0, 0, 13, 0, 0]))

# The parse function converts a command string into tokens and dispatches game functions
# to enact these commands. For example, if the player enters the command
# "deposit coin" then this function will convert that into two words
# "DEPOS" and "COIN." These tokens can be looked up in the Things and Verbs
# dictionaries above.

# This function does some prescreening to make sure that the Verbs and
# Things exist in-game and that the Things are accessible to the Player,
# in the inventory or in the current room.

def parse(command):
	# Convert to uppercase and split words into list.
	words = command.upper().split()[:2]

	# Return now if the command was blank.
	if len(words) <= 0:
		return

	# Split command into two words and shorten to five characters.
	if len(words) > 0:
		verb = words[0][:5]

	# Commands might have a verb only and not a noun. E.g., "LOOK"
	if len(words) > 1:
		noun = words[1][:5]
	else:
		noun = None

	# Directions get special treatment as they can be verbs ("NORTH") or an nouns ("GO NORTH")
	if verb in Directions:				
		doGo(verb)
	elif verb == "GO":
		doGo(noun[0])
	elif not verb in Verbs:
		# Did the player enter a verb found in the Verbs dictionary?

		print("I don't know the verb", verb)
	else:
		# If the player entered a legal Verb, find the associated function
		act = Verbs[verb]

		# If this is a verb that needs no noun, go ahead and call it. E.g., "LOOK"
		if noun == None:
			act(None)
		else:
			if not noun in Things:
				# Did the player enter a noun found in the game?

				print("I don't know the object", noun)
			else:
				# If this verb requires a noun, then make sure that the noun is accesible
				# to the player. The steps here are:
				# 1. Look up the noun in the Things table.
				# 2. Is the Thing either in the current room or in the player's inventory.
				# 3. If not, say that the Thing isn't here.
				# 4. If yes, then call the verb's function with the Thing: act(thing).

				# Again, the TYPE verb is a special case for typing words into the 
				# in-game computer.

				thing = Things[noun]
				if verb != "TYPE" and not thing in Rooms[player.room].contents and not thing in player.inventory:
					print("It isn't here.")
				else:
					return act(thing)

# The functions below enact the game code for the verbs, things and rooms above.
# Drawing on the original BASIC source, many of these refer to room numbers, 
# which correlate to the game map in the Rooms list above. In general, these
# functions all work like this:

# 1. Is the given object appropriate to the verb? E.g., the player cannot
# GO CANDY BAR as if it were a direction. The player also cannot EAT COIN
# or EAT WEST as if a direction were a noun.

# 2. Perform appropriate manipulations on the game state. E.g., the command
# EAT CANDY should destroy the CANDY object so it can't be used again.
# E.g., the TAKE command should remove an object from the room contents
# and add it to the player's inventory list.

# Again, the TYPE command is complex as it manipulates an in-game computer.

# 3. Print appropriate success and failure messages.

# 4. Return True or False based on the completion (or not) of the command.
# This is mostly relevant to counting completed turns while the computer 
# is running. If too many turns pass, the computer dies.

def doGo(object):
	if Rooms[player.room].exits[Directions[object][0]] == 0:
		print("You can't go that way.")
	elif player.room == 4 and not player.ratsAsleep:
		print("The rats look too fierce.")
	else:
		player.room = Rooms[player.room].exits[Directions[object][0]]
		player.wantDescription = True

	return True

def doEat(object):
	if not object == Things["CANDY"]:
		print("That's silly!")
	else:
		print("GAG! COUGH! CHOKE! PUKE!")
		if Things["CANDY"] in Rooms[player.room].contents:
			Rooms[player.room].contents.remove(Things["CANDY"])
		if Things["CANDY"] in player.inventory:
			player.inventory.remove(Things["CANDY"])

def doKick(object):
	if object == None:
		print("Please give an object.")
		return False
	elif object != Things["COMPU"] or player.computerDead:
		print("Nothing happens.")
	elif player.computerActive:
		player.counter = 9
	else:
		print("The computer starts up!")
		print("The console displays: 'PLEASE LOG IN.'")
		player.computerActive = -1

	return True

def doInsert(object):
	if object == None:
		print("Please give an object.")
		return False
	elif object != Things["COIN"]:
		print("That's silly!")
		return False
	elif player.room != 6:
		print("You can't do that now.")
	else:
		player.inventory.remove(Things["COIN"])
		Rooms[player.room].contents.append(Things["CANDY"])
		print("A candy bar comes out.")

	return True

def doType(object):
	if object == None:
		print("Please give an object.")
		return False
	elif player.room != 10:
		print("You can't do that now.")		
	elif not player.computerActive:
		print("The computer isn't running.")
	elif not player.loggedIn:
		if object != Things["ROAD"]:
			print("'Invalid login ID.'")
		else:
			print("'ROAD' logged in.'")
			player.loggedIn = True
	elif not player.copyRunning:
			if object == Things["DIR"]:
					print("'COPY LOGOU ADVEN'")
			elif object == Things["ADVEN"]:
					print("'Welcome to Adventure! W#uld Y#$*'")
					player.counter = 9
			elif object == Things["COPY"]:
					print("Mount tape then type filename.")
					player.copyRunning = True 
			else:
				print("'Invalid command.'")
	else:
		player.copyRunning = False
		if player.mountedTape:
			if object in [Things["COPY"], Things["DIR"], Things["ADVEN"]]:
				print("The tape spins...")
				print("'File copied.'")

				if object == Things["ADVEN"]:
						print("Congratulations, you've done it!")
						exit()
			else:
				print("'No such file'")
		else:
				print("'Error: Tape not mounted'")

	return True

def doTake(object):
	if object == None:
		print("Please give an object.")
		return False
	elif not object.takable:
		print("That's beyond your ability.")
		return True
	elif object in player.inventory:
		print("You already have it!")
		return False
	elif object == Things["TERMI"] and not player.coinFound:
		print("There was a coin underr it.")
		Rooms[player.room].contents.append(Things["COIN"])
		player.coinFound = True
	elif object == Things["TAPE"]:
		mountedTape = False

	Rooms[player.room].contents.remove(object)
	player.inventory.append(object)
	print("O.K.")

	return True

def doDrop(object):
	if object == None:
		print("Please give an object.")
		return False
	elif not object in player.inventory:
		print("You don't have it.")
	else:
		player.inventory.remove(object)
		Rooms[player.room].contents.append(object)
		if player.room == 4 and object == Things["CANDY"]:
			print("The rats devour the candy and get sleepy.")
			Rooms[player.room].contents.remove(Things["CANDY"])
			Things["RATS"].description = "Sleepy rats"
			player.ratsAsleep = True

	return True

def doInventory(object):
	print("You are carrying:")
	if player.inventory == []:
		print("Nothing.")
	else:
		for thing in player.inventory:
			print(thing.description)

	return True

def doMount(object):
	if object == None:
		print("Please give an object.")
		return False
	elif object != Things["TAPE"]:
		print("That's silly!")
		return False
	elif player.room != 10 or player.mountedTape:
		print("You can't do that now.")
	else:
		player.inventory.remove(Things["TAPE"])
		Rooms[player.room].contents.append(Things["TAPE"])
		player.mountedTape = True
		print("O.K.")

	return True

def doRead(object):
	if object == None:
		print("Please give an object.")
		return False
	elif object == Things["MACHI"]:
		print("'INSERT COIN.'")
	elif object == Things["MANUA"]:
		print("'...USER ID IS ROAD...'")
		print("'TYPE DIR FOR LIST OF COMMANDS...'")
		print("The rest is illegible.")
	else:
		print("Nothing is written on it.")

	return True

def doFight(object):
	print("That won't work.")
	return True

def doStart(object):
	if object == None:
		print("Please give an object.")
		return False
	elif object == Things["COMPU"]:
		print("Nothing happens.")
	else: 
		print("That's silly!")
		return False

	return True

def doOpen(object):
	if object == None:
		print("Please give an object.")
		return False
	elif object == Things["DESK"]:
			if player.manualFound:
				print("It already is.")
			else:
				Rooms[player.room].contents.append(Things["MANUA"])
				player.manualFound = True
				print("Inside it is a manual.")
	else:
		print("That's silly!")
		return False

	return True

def doQuit(object):
	exit()

def doLook(object):
	player.wantDescription = True
	return False

def doWind(object):
	if object == Things["CLOCK"]:
		if player.bats:
			print("The clock chimes deafeningly and something flies past.")
			player.bats = False
			Rooms[13].contents.remove(Things["BATS"])
		else:
			print("It's fully wound.")
	else:
		print("That's silly!")
	
	return True

def doExamine(object):
	if object == Things["MACHI"] or object == Things["MANUA"]:
		print("Something is written there.")
	elif object == Things["DESK"] and not player.manualFound:
		print("It is closed.")
	elif object == Things["TERMI"]:
		print("It looks beyond repair.")
	elif object == Things["COMPU"]:
		print("This is an ancient mainframe with a console.")
	elif object == Things["CLOCK"]:
		print("There is a large handle for winding the clock.")
	elif object == Things["TAPE"] and player.mountedTape:
		print("It is mounted on the computer.")
	else:
		print("You see nothing special.")

	return True

# The Verbs dictionary correlates tokens with the functions above. This 
# structure allwos for synonyms to be used. E.g., GET and TAKE call the
# same function above.

Verbs = {
	"N": doGo,
	"S": doGo,
	"E": doGo,
	"W": doGo,
	"U": doGo,
	"D": doGo,
	"GO": doGo,
	"EAT": doEat,
	"KICK": doKick,
	"INSER": doInsert,
	"DEPOS": doInsert,
	"TYPE": doType,
	"TAKE": doTake,
	"GET": doTake,
	"DROP": doDrop,
	"THROW": doDrop,
	"INVEN": doInventory,
	"I": doInventory,
	"MOUNT": doMount,
	"READ": doRead,
	"FIGHT": doFight,
	"KILL": doFight,
	"START": doStart,
	"POWER": doStart,
	"OPEN": doOpen,
	"QUIT": doQuit,
	"LOOK": doLook,
	"WIND": doWind,
	"EXAMI": doExamine
}

# Create the player object.
player = Player()

# MAIN GAME LOOP - There are only a few general steps in the loop:
# 1. Display the current room description (if needed).
# 2. Display a prompt and read the player's command: input("> ")
# 3. Parse the given command and record whether the action can be completed.

# The other steps after are mechanics specific to this game:
# 1. If the computer is active, then the player only gets 10 turns to finish
# using the computer. Beyond this time limit, the computer dies and the game
# cannot be completed.

# 2. If the player enters a room with Bats in it, then the player is returned
# to the beginning of the game (player.room = 1).

while True:
	if player.wantDescription:
		Rooms[player.room].describe()

	player.wantDescription = False

	complete = parse(input("> "))

	if complete:
		if player.computerActive:
			player.counter = player.counter + 1
			if player.counter >= 10 and player.room == 10:
				print("The computer dies with a loud pop.")
				player.computerDead = True
				player.computerActive = False
				Things["COMPU"].description = "Dead computer"

	if Things["BATS"] in Rooms[player.room].contents:
		print("A horde of bats carries you out.")
		player.room = 1
		player.wantDescription = False
